#include <openssl/ssl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/timeb.h>

#include "openssl/bn.h"
#include <gmp.h>
#include "openssl/rsa.h"
#include "openssl/dsa.h"
#include "openssl/dh.h"

#include "parameters.h"
#include "tools.h"
#include "psi_cli.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h> 


/* global variables */
int n_c, n_s,sockfd;
RSA *params;
gRSA rsa;
mpz_t *inv;
char **tc;
int pipe1[2], pipe2[2], pipe3[2];
int new_pipe[2];
double *times;
struct timeb *tstart,*tend;


int main(int argc, char *argv[])
{ 
    struct sockaddr_in serv_addr;
    struct hostent *server;
    int i;
    FILE *fp_out=fopen("times_cli.res", "a"); //Will write benchmark
    pthread_t t1, t2, t3, t4;
    pthread_t new_t1, new_t2;
    pthread_t final_t1;
    struct timeb tstart_tot,tend_tot;
    
    /* Read arguments from command line */
	if(argc != 5) {
		printf("ERROR\n");
		printf("Usage ./psi_cli n_s n_c srv_host_name srv_port\n");
		exit(-1);
	} 

    /* Declaring variables */
    int n_tim=7;
    times=(double *) malloc(n_tim*sizeof(double));
    for(i=0;i<n_tim;i++) times[i]=0.0;
    tstart=(struct timeb *) malloc(n_tim*sizeof(struct timeb));
    tend=(struct timeb *) malloc(n_tim*sizeof(struct timeb));
	double t_tot=0.0;

    n_s=atoi(argv[1]); //size of server set
    n_c=atoi(argv[2]); //size of client set

	/* Initalizing client socket */
	sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if (sockfd < 0) 
		printf("ERROR opening socket\n");
	server = gethostbyname(argv[3]);
	if (server == NULL) {
		printf("ERROR, no such host\n");
		exit(0);
	}
	bzero((char *) &serv_addr, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	bcopy((char *)server->h_addr,(char *)&serv_addr.sin_addr.s_addr,server->h_length);
	serv_addr.sin_port = htons(atoi(argv[4]));
 
    /* connect to server socket */
	if (connect(sockfd,(struct sockaddr *) &serv_addr,sizeof(serv_addr)) < 0){ 
		printf("ERROR connecting\n"); exit(-1);}


    inv=malloc(n_c*sizeof(mpz_t));
    tc=(char **) malloc(n_c*sizeof(char *));


    if(udbg)
        printf("STARTING %d\n", (int)time(NULL));

    M_Time(START, &tstart_tot, &tend_tot); //Start Counting Total Time

	/* Load server's public key */
	params= RSA_new();
	load_pub_key(params,&rsa);
    
    /* Initalize pipes */
    pipe(pipe1);
    pipe(pipe2);
    pipe(pipe3);

    /* Start thread */
    pthread_create(&t1, NULL, (void *)read_ci, NULL); //read c_i-s
    pthread_create(&t2, NULL, (void *)compute_mui, NULL); //compute mu_i-s, i.e., H(c_i)*r_i^e
    pthread_create(&t3, NULL, (void *)send_mui, NULL); //send mu_i-s
    pthread_create(&t4, NULL, (void *)invert_ri, NULL); //compute 1/r_i

    pthread_join(t1, NULL); //wait for thread to end
    pthread_join(t2, NULL); //wait for thread to end
    pthread_join(t3, NULL); //wait for thread to end
    pthread_join(t4, NULL); //wait for thread to end
    

    pipe(new_pipe);
    pthread_create(&new_t1, NULL, (void *)receive_mupi, NULL); //receive mup_i-s
    pthread_create(&new_t2, NULL, (void *)process_mupi, NULL); //process mup_i-s, i.e., tc_i=mup_i/r_i
    pthread_join(new_t1, NULL); //wait for thread
    pthread_join(new_t2, NULL); //wait for thread
    shutdown(sockfd,SHUT_WR); //close the socket

    pthread_create(&final_t1, NULL, (void *)finalize, NULL); //receive ts_j-s and compute intersection
    pthread_join(final_t1, NULL); //wait for thread

    shutdown(sockfd,SHUT_RD); //close the socket
    t_tot+=M_Time(STOP, &tstart_tot, &tend_tot); //Stop Counting Total Time

    fprintf(fp_out, "%d %d ", n_s, n_c);
    if(udbg)
        printf("|Init\t|Send\t|Cli1\t|Invert\t|Recv\t|Proc\t|Final\t|TOT\n");
    for(i=0;i<n_tim;i++) {
        if(udbg)
            printf("|%.2f\t", times[i]);
        fprintf(fp_out, "%.2f ", times[i]);
    }
    if(udbg) printf("|%.2f\n", t_tot);
    fprintf(fp_out, "%.2f\n", t_tot);

    return 1;
}


void send_mui() 
{
    int i; 
    char sock_buffer[rsa_buff+2];


    for (i=0; i<n_c;i++) {
        bzero(sock_buffer,rsa_buff+2);
        read_from_pipe(pipe2[0], sock_buffer, rsa_buff); //read from pipe
        M_Time(START, &(tstart[1]),&(tend[1])); //times[1] send time
	    if(!write_to_socket(sockfd, strlen(sock_buffer), sock_buffer))
        {
            printf("Error writing to socket\n", i); exit(-1);
        }
        times[1]+=M_Time(STOP, &(tstart[1]),&(tend[1])); //times[1] send time
    }
    
    if(udbg)
        printf("Send_mui in %.2f ms, time %d\n", times[1], (int)time(NULL));


}

void read_ci()
{
 FILE *fp_cl=fopen("client.dat","r"); //read client items from file
 char sock_buffer[rsa_buff+1], fdh[rsa_buff+1];
    int i; //, pipe1=(pipe[1]);

 for (i=0; i<n_c;i++) {
     M_Time(START, &(tstart[0]), &(tend[0])); //times[0] init time
     bzero(sock_buffer, rsa_buff+1);
     bzero(fdh, rsa_buff+1);
     fscanf(fp_cl,"%s",sock_buffer); //read
     FDH_str_to_str(sock_buffer, strlen(sock_buffer), fdh); //compute hash
     times[0]+=M_Time(STOP, &(tstart[0]),&(tend[0])); //times[0] init time
     write_to_pipe(pipe1[1], fdh, strlen(fdh)); //write mu[i] to the pipe
 }
    if(udbg)
        printf("Read_ci in %.2f ms, time %d\n", times[0], (int)time(NULL));

 fclose(fp_cl);
}


void compute_mui() {
    int i;
    mpz_t ri, mui, hci, temp;
    BIGNUM *rand=BN_new();
    char sock_buffer[rsa_buff+2], fdh[rsa_buff+1];
    

    for (i=0; i<n_c;i++) {
        bzero(sock_buffer, rsa_buff+2);
        bzero(fdh, rsa_buff+1);
        mpz_init(ri);
        mpz_init(mui);
        mpz_init(temp);	
        mpz_init(hci);	
        M_Time(START, &(tstart[2]),&(tend[2])); //times[2] compute mui
        rand_num(rand,params->n);//pick a random in Z_N
        BN2mpz(ri,rand);//transform it to mpz_t
        mpz_powm(temp,ri,rsa.e,rsa.n);//temp=r^e mod N
        times[2]+=M_Time(STOP, &(tstart[2]),&(tend[2])); //times[1] compute mui
        read_from_pipe(pipe1[0], fdh, rsa_buff); //read from pipe
        mpz_set_str(hci, fdh, 62);
        M_Time(START, &(tstart[2]),&(tend[2])); //times[2] compute mui
        mpz_mul(mui,hci,temp);
        mpz_mod(mui,mui,rsa.n);//mu[i]=FDH(c_i)*r^e mod N
        times[2]+=M_Time(STOP, &(tstart[2]),&(tend[2])); //times[1] compute mui
        mpz_get_str(sock_buffer, 62, mui);
        write_to_pipe(pipe2[1], sock_buffer, strlen(sock_buffer)); //write mu[i] to the pipe
        bzero(sock_buffer, rsa_buff+2);
        mpz_get_str(sock_buffer, 62, ri);
        write_to_pipe(pipe3[1], sock_buffer, strlen(sock_buffer)); //write r[i] to the pipe
    }
    if(udbg)
        printf("Compute_Mui in %.2f ms, time %d\n", times[2], (int)time(NULL));

}


void invert_ri() {
    int i; 
    mpz_t ri;
    char r_buff[rsa_buff+1];

    for (i=0;i<n_c;i++) {
        mpz_init(ri);
        mpz_init(inv[i]);
        bzero(r_buff, rsa_buff+1);
        read_from_pipe(pipe3[0], r_buff, rsa_buff); //read from pipe
        M_Time(START, &(tstart[3]),&(tend[3]));
        mpz_set_str(ri, r_buff, 62);
        mpz_invert(inv[i], ri, rsa.n);
        times[3]+=M_Time(STOP, &(tstart[3]),&(tend[3])); //times[3] invert_ri
    }
    if(udbg)
        printf("Invert_ri in %.2f ms, time %d\n", times[3], (int)time(NULL));
}


void receive_mupi(){
    int i;
    char sock_buffer[rsa_buff+1];
    mpz_t mupi;

    for (i=0;i<n_c;i++) {
        M_Time(START, &(tstart[4]),&(tend[4]));
        bzero(sock_buffer,rsa_buff+1);
        if(!read_from_socket(sockfd,rsa_buff,sock_buffer)) {
            printf("Error reading from socket\n");
            exit(-1);
        }
        mpz_init(mupi);
        mpz_set_str(mupi, sock_buffer, 62);
        times[4]+=M_Time(STOP, &(tstart[4]),&(tend[4])); //times[4] time to receive mupi
        write_to_pipe(new_pipe[1], sock_buffer, strlen(sock_buffer)); //write mup[i] to the pipe
    }
    if(udbg)
        printf("Receive_mupi in %.2f ms, time %d\n", times[4], (int)time(NULL));

}

void process_mupi() {
    int i; 
    char buffer[rsa_buff+1];
    mpz_t mupi, temp;

    for (i=0;i<n_c;i++) {
        mpz_init(mupi);
        bzero(buffer,rsa_buff+1);
        read_from_pipe(new_pipe[0], buffer, rsa_buff); //read from pipe
        M_Time(START, &(tstart[5]),&(tend[5]));
        mpz_set_str(mupi, buffer, 62);       
        mpz_init(temp);
        mpz_mul(temp,mupi,inv[i]);
        mpz_mod(temp,temp,rsa.n); //temp=mup[i]/r[i] mod N
        tc[i]=malloc(hash_buff);
        hash_mpz_to_str(temp, tc[i]); //tc[i]=H(mup[i]/r[i])
        times[5]+=M_Time(STOP, &(tstart[5]),&(tend[5])); //times[5] time to process mupi

    }
    if(udbg)
        printf("Process_mupi in %.2f ms, time %d\n", times[5], (int)time(NULL));

}

void finalize() {
    int i,j;
    char sock_buffer[hash_buff+1], cmd[64];

    M_Time(START, &(tstart[6]),&(tend[6]));
    for (j=0;j<n_s;j++) {
        bzero(sock_buffer,hash_buff+1);
        if(!read_from_socket(sockfd,hash_buff,sock_buffer)) {
            printf("Error reading from socket\n");
            exit(-1);
        }
        
        if (udbg) {
        /* TO DO: Replace this search with hash tables and improve running times for large sets!!! */
            for(i=0;i<n_c;i++) {
                if(!strcmp(tc[i],sock_buffer)) {
                    printf("In intersection: c[%d] ",i);
                    fflush(stdout);
                    sprintf(cmd,"awk 'NR==%d{print;exit}' client.dat",i+1);
                    system(cmd);
                }
            }
        }
    }
    times[6]+=M_Time(STOP, &(tstart[6]),&(tend[6])); //times[6] to finalize
    if(udbg)
        printf("Finalize in %.2f ms, time %d\n", times[6], (int)time(NULL));

}
    
