#include <openssl/ssl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/timeb.h>

#include <gmp.h>
#include "openssl/bn.h"
#include "openssl/rsa.h"
#include "openssl/dsa.h"
#include "openssl/dh.h"

#include "parameters.h"
#include "tools.h"

#include <unistd.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <pthread.h>

/* global variables */
int sockfd, newsockfd, n_s, n_c;
RSA *params;
gRSA rsa;
int pipe1[2], pipe2[2], pipe3[2];
char **ts;
mpz_t *mup;

double *times;
struct timeb *tstart,*tend;

/* server subroutines */
void receive_from_client();
void blind_sign();
void send_bsig();
void compute_tags();
void send_tags();



int main(int argc, char *argv[])
{ 
	int portno;
	socklen_t clilen;
	int optval=1;
	int optlen;
	struct sockaddr_in serv_addr, cli_addr;
	int i, j;

    /* Timing */
	FILE *fp_out=fopen("times_srv.res", "a");
    struct timeb tstart_tot,tend_tot;

	if(argc != 4) {
		printf("ERROR\n");
		printf("Usage ./psi_srv n_s n_c PORT\n");
		exit(-1);
	}
	n_s=atoi(argv[1]);	
	n_c=atoi(argv[2]);
	portno=atoi(argv[3]);
    pthread_t t1, t2, t3, t4, t5;

    /* Declaring variables for timing */
    int n_tim=5;
    times=(double *) malloc(n_tim*sizeof(double));
    for(i=0;i<n_tim;i++) times[i]=0.0;
    tstart=(struct timeb *) malloc(n_tim*sizeof(struct timeb));
    tend=(struct timeb *) malloc(n_tim*sizeof(struct timeb));
	double t_tot=0.0;

    	    
	/* Init Server Socket */
	sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if (sockfd < 0) {
	        printf("ERROR opening socket\n"); exit(-1);}
	bzero((char *) &serv_addr, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = INADDR_ANY;
	serv_addr.sin_port = htons(portno);
	setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof optval);
	if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
		printf("ERROR on binding\n");
		exit(-1);
	}
	listen(sockfd,5);


    /* Load RSA keys */
	params= RSA_new();
	load_keys(params,&rsa);

    pipe(pipe1);

    
	/* Server listens to the socket (Must ignore waiting time) */
	clilen = sizeof(cli_addr);
	newsockfd = accept(sockfd,(struct sockaddr *) &cli_addr,&clilen);
	if (newsockfd < 0) { 
          printf("ERROR on accept\n");
	  exit(-1);
	}
    //Server has accepted connection from client
	shutdown(sockfd,2);
    
    if(udbg)
        printf("STARTING %d\n", (int)time(NULL));
    M_Time(START, &tstart_tot, &tend_tot); //Start Counting Total Time


    /* Compute Server's "Tags" */
    ts=(char **)malloc(n_s*sizeof(char*)); //Init memory for server's tags
    pthread_create(&t4, NULL, (void *)compute_tags, NULL); //compute ts_j=H'(H(s_j^d))
    
    /* Receive mu_i-s from client and send back mu'_i-s */
    mup=(mpz_t *)malloc(n_c*sizeof(mpz_t));
    pthread_create(&t1, NULL, (void *)receive_from_client, NULL); //receive mu_i=H(c_i)*r_i^e
    pthread_create(&t2, NULL, (void *)blind_sign, NULL); //compute mup_i=mu_i^d
    pthread_join(t1, NULL); //wait for thread to end
    pthread_join(t2, NULL); //wait for thread to end
    pthread_create(&t3, NULL, (void *)send_bsig, NULL); //send mup_i-s
    pthread_join(t3, NULL); //wait for thread to end
    pthread_join(t4, NULL); //wait for thread to end
    
    /* Send ts_j-s */
    send_tags();
    
    fprintf(fp_out, "%d %d ", n_s, n_c);

    if(udbg)
        printf("|Recv\t|BSig\t|Send\t|Tags\t|SndTgs\t|TOT\n");
    for(i=0;i<n_tim;i++) {
        if(udbg) printf("|%.2f\t", times[i]);
        fprintf(fp_out, "%.2f ", times[i]);
    }
 
   t_tot+=M_Time(STOP, &tstart_tot, &tend_tot); //Stop Counting Total Time
	shutdown(newsockfd,0);
    //close(newsockfd);
    //close(sockfd);

    if (udbg)
        printf("|%.2f\n", t_tot);
    fprintf(fp_out, "%.2f\n", t_tot);

}

void receive_from_client(){
    int i;
    char sock_buffer[rsa_buff+1];
    
    for(i=0;i<n_c;i++) {
        //Read mu[i] from the socket
        M_Time(START, &(tstart[0]), &(tend[0])); 
        bzero(sock_buffer,rsa_buff+1);
        if(!read_from_socket(newsockfd,rsa_buff,sock_buffer)) {
            printf("Error reading from socket\n"); exit(-1);
        }
        times[0]+=M_Time(STOP, &(tstart[0]),&(tend[0]));
        write_to_pipe(pipe1[1], sock_buffer, strlen(sock_buffer)); //write mu_i to the pipe
    }
    if (udbg) printf("Receive_from_client %.2f done %d\n", times[0],(int)time(NULL));
}

void blind_sign() {
    int i;
    mpz_t mui, mupi;
    char buffer[rsa_buff+1];

    for(i=0;i<n_c;i++) {
        mpz_init(mup[i]);
        mpz_init(mui);
        bzero(buffer,rsa_buff+1);
        read_from_pipe(pipe1[0], buffer, rsa_buff); //read from pipe
        M_Time(START, &(tstart[1]), &(tend[1]));
        mpz_set_str(mui, buffer, 62);
        sign_crt_gmp(mup[i],mui,&rsa); //Blind sign mup_i=mu_i^d
        times[1]+=M_Time(STOP, &(tstart[1]),&(tend[1]));
    }    
    if (udbg) printf("Blind_Sign %.2f done %d\n", times[1], (int)time(NULL));

}

void send_bsig(){
    int i;
    char sock_buffer[rsa_buff+1];

    M_Time(START, &(tstart[2]), &(tend[2])); 
    for(i=0;i<n_c;i++) {
        bzero(sock_buffer,rsa_buff+1);
        mpz_get_str(sock_buffer, 62, mup[i]);

        //Send mu'_i
        if(!write_to_socket(newsockfd, strlen(sock_buffer), sock_buffer))
        {
           printf("Error writing to socket\n"); exit(-1);
        }
    }
    times[2]+=M_Time(STOP, &(tstart[2]),&(tend[2]));
    if (udbg) printf("Send_bsig %.2f done %d\n", times[2], (int)time(NULL));

}

void compute_tags() {
    FILE *fp_dat=fopen("server.dat","r");
    int j;
    mpz_t temp;
    mpz_t hsj;
    char buffer[max_size];
    
    M_Time(START, &(tstart[3]), &(tend[3])); 
    for (j=0; j<n_s;j++) {
        fflush(stdout);
        fscanf(fp_dat,"%s",&buffer);
        mpz_init(hsj);
        FDH_str_to_mpz(buffer, strlen(buffer), &hsj); //hs_j=H(s_j)
        mpz_init(temp);
        sign_crt_gmp(temp,hsj,&rsa);//sign hs_j^d
        *(ts+j)=(char *)malloc((hash_buff+1)*sizeof(char));
        hash_mpz_to_str(temp, (ts[j]));//tag ts_j=H'(hs_j^d)
    }
    times[3]+=M_Time(STOP, &(tstart[3]),&(tend[3]));

    if (udbg) printf("Compute_tags %.2f done %d\n", times[3], (int)time(NULL));

}

void send_tags() {
    int j;
    M_Time(START, &(tstart[4]), &(tend[4]));

    for (j=0; j<n_s;j++) {
        // Send ts_j
        if(!write_to_socket(newsockfd, strlen(ts[j]), ts[j]))
        { printf("Error writing to socket\n"); exit(-1); }
    }
    times[4]+=M_Time(STOP, &(tstart[4]),&(tend[4]));
    if (udbg) printf("Send_tags %.2f done %d\n", times[4], (int)time(NULL));

}
