#include "tools.h"


void load_pub_key(RSA *params, gRSA *rsa) {
	unsigned char buffer[1024];
	FILE *fp=fopen("rsa.pub", "r");
	fscanf(fp, "%s", &buffer);
	BN_hex2bn(&(params->n), buffer);
	mpz_init(rsa->n);
	BN2mpz(rsa->n,params->n);
	fscanf(fp, "%s", &buffer);
	BN_hex2bn(&(params->e), buffer);
	mpz_init(rsa->e);
	BN2mpz(rsa->e,params->e);
	fclose(fp);
}

void BN2mpz(mpz_t s, const BIGNUM *bn)
{
    size_t len;
    void *p;
    len = BN_num_bytes(bn);
    p = malloc(len);
    BN_bn2bin(bn, p);
    mpz_import(s, len, 1, 1, 1, 0, p);
    free(p);
}



void load_keys(RSA *params, gRSA *rsa) {
	unsigned char buffer[1024];
	FILE *fp=fopen("rsa.pub", "r");
	fscanf(fp, "%s", buffer);
	mpz_init(rsa->n);
    mpz_set_str(rsa->n,buffer,16);
	
	BN_hex2bn(&(params->e), buffer);
	mpz_init(rsa->e);
	BN2mpz(rsa->e,params->e);
	fclose(fp);
	fp=fopen("rsa.prv", "r");
	fscanf(fp, "%s", &buffer);
	BN_hex2bn(&(params->d), buffer);
	mpz_init(rsa->d);
	BN2mpz(rsa->d,params->d);
	fscanf(fp, "%s", &buffer);
	BN_hex2bn(&(params->p), buffer);
	mpz_init(rsa->p);
	BN2mpz(rsa->p,params->p);
	fscanf(fp, "%s", &buffer);
	BN_hex2bn(&(params->q), buffer);
	mpz_init(rsa->q);
	BN2mpz(rsa->q,params->q);
	fscanf(fp, "%s", &buffer);
	BN_hex2bn(&(params->dmp1), buffer);
	mpz_init(rsa->dmp1);
	BN2mpz(rsa->dmp1,params->dmp1);
	fscanf(fp, "%s", &buffer);
	BN_hex2bn(&(params->dmq1), buffer);
	mpz_init(rsa->dmq1);
	BN2mpz(rsa->dmq1,params->dmq1);
	fscanf(fp, "%s", &buffer);
	BN_hex2bn(&(params->iqmp), buffer);
	mpz_init(rsa->iqmp);
	BN2mpz(rsa->iqmp,params->iqmp);
	fclose(fp);
}

void FDH(BIGNUM *fdh, BIGNUM *tohash) {
	int i=0, cycles=rsa_key_length/(SHA_DIGEST_LENGTH*8)+1;
	char buff[rsa_key_length/8*2];
	char *ibuf=malloc(BN_num_bytes(tohash)+1), cat[8];
    unsigned char *obuf;
    
	sprintf(buff,"");
	for(i=0;i<cycles;i++) {
		ibuf = BN_bn2dec(tohash);
		sprintf(cat,"||%d",i);
		strcat(ibuf,cat);
        obuf=SHA1(ibuf, strlen(ibuf), NULL);
		strcat(buff,obuf);
	}
	buff[rsa_key_length/8]='\0';
	BN_bin2bn(buff, strlen(buff), fdh);
}

void FDH_str_to_mpz(const unsigned char *input, int len, mpz_t *output){

    int i, nbits=rsa_key_length;
    int repetition = (nbits + hash_size - 1) / hash_size;
    unsigned char out[rsa_key_length/8*2], ibuf[len*2];
    sprintf(out,"");
    
    for (i = 0; i < repetition; i++)
    {
        sprintf(ibuf,"%s||%d",input,i);
        strcat(out, SHA1((const unsigned char*)ibuf, strlen(ibuf), NULL));
    }
    mpz_import(*output, strlen(out), 1, sizeof(char), 0, 0, (const unsigned char *)out);
}

void FDH_str_to_str(const unsigned char *input, int len, char *output){
    
    int i, nbits=rsa_key_length;
    int repetition = (nbits + hash_size - 1) / hash_size;
    unsigned char out[rsa_key_length/8*2], ibuf[len*2];
    sprintf(out,"");
    mpz_t temp;
    
    mpz_init(temp);
    for (i = 0; i < repetition; i++)
    {
        sprintf(ibuf,"%s||%d",input,i);
        strcat(out, SHA1((const unsigned char*)ibuf, strlen(ibuf), NULL));
    }
    mpz_import(temp, strlen(out), 1, sizeof(char), 0, 0, (const unsigned char *)out);
    mpz_get_str(output, 62, temp); 
}


int sign_crt_gmp(mpz_t *sig, mpz_t *msg, gRSA *rsa) {
	mpz_t m1, m2, tmp, hq;
	mpz_init(m1);mpz_init(m2);mpz_init(tmp);mpz_init(hq);
	mpz_powm(m1, msg, rsa->dmp1, rsa->p);
	mpz_powm(m2, msg, rsa->dmq1, rsa->q);
	mpz_add(tmp, m1, rsa->p);
	mpz_sub(tmp, tmp, m2);
	mpz_mod(tmp, tmp, rsa->p);
	mpz_mul(hq, rsa->iqmp, tmp);
	mpz_mod(hq, hq, rsa->p);
	mpz_mul(hq, hq, rsa->q);
	mpz_add(sig, m2, hq);
}

void rand_num(BIGNUM *x, BIGNUM *n)
{
    //BIGNUM *gcd=BN_new();
    
    //BN_CTX *ctx=BN_CTX_new();
    
    BN_rand_range(x, n); //choose a random number x \in [1, n-1]
    //BN_gcd(gcd, x, n, ctx);
    /*while (BN_cmp(gcd, BN_value_one())) //is gcd == 1? 
    {
        BN_rand_range(x, n);
        BN_gcd(gcd, x, n, ctx);
    }*/
    //BN_clear_free(gcd); //free gcd and overwrite the structure
}

double F_Time(int s)
{
    double measure;
    static struct timeb tstart,tend;
    long tdiff;
    
    if (s==START) {
        ftime(&tstart);
        return(0);
    }
    else {
        ftime(&tend);
        tdiff=(long)tend.millitm-(long)tstart.millitm;
        measure=((double)(tend.time-tstart.time))+((double)tdiff)/1000.0;
        if (measure < 0.00000001) measure=0.00000001;
        
        float t = measure*1000.0;
        
        return t;
    }
}

double M_Time(int s, struct timeb *tstart, struct timeb *tend)
{
    double measure;
    //static struct timeb tstart,tend;
    long tdiff;
    
    if (s==START) {
        ftime(tstart);
        return(0);
    }
    else {
        ftime(tend);
        tdiff=(long)tend->millitm-(long)tstart->millitm;
        measure=((double)(tend->time-tstart->time))+((double)tdiff)/1000.0;
        if (measure < 0.00000001) measure=0.00000001;
        
        float t = measure*1000.0;
        
        return t;
    }
}



void hash_mpz_to_str(mpz_t *input, char *out){
    
    unsigned char ibuf[rsa_buff*2], obuf[20], prep[hash_buff+1];
    mpz_t t;
    mpz_get_str(ibuf, 16, *input);
    SHA1(ibuf, strlen(ibuf), obuf);
    mpz_init(t);
    mpz_import(t, 20, 1, sizeof(char), 0, 0, obuf);
    mpz_get_str(out, 62, t);
    if(strlen(out)<hash_buff) {
        sprintf(prep,"0%s",out);
        strcpy(out,prep);
    }
    if(strlen(out)<hash_buff) {
        sprintf(prep,"0%s",out);
        strcpy(out,prep);
    }

}


int read_from_socket(int socket, int max_len, char *output) {
	char small[2];
	int cnt=0, sock_n;
	sprintf(output, "");
	while(1) {
	    bzero(small,2);
	    sock_n= read(socket,small,1);
	    if (sock_n < 0) { printf("ERROR reading from socket\n"); return(0);}
	    cnt+=sock_n;
	    small[1]='\0';
	    if(small[0]=='@') {
		output[strlen(output)]='\0';
		return strlen(output);
	    }
	    strcat(output, small);
	}
	return 0;
}

int write_to_socket(int socket, int len, char *input) {
	int sock_n;
	char buff[strlen(input)+1];
	sprintf(buff, "%s@", input);
    sock_n=write(socket,buff,strlen(buff)); //send mu[i] on socket
	if(sock_n<0) return 0;
	else return 1;
}

int write_to_pipe(int pipe, char *input, int len) {
	int pipe_n;
	char buff[strlen(input)+1];
	sprintf(buff, "%s@", input);
    pipe_n=write(pipe,buff,strlen(buff)); //send mu[i] on socket
	if(pipe_n<0) return 0;
	else return 1;
}


int read_from_pipe(int pipe, char *output, int max_len) {
	char small[2];
	int cnt=0, pipe_n;
	sprintf(output, "");
	while(1) {
	    bzero(small,2);
	    pipe_n= read(pipe,small,1);
	    if (pipe_n < 0) { printf("ERROR reading from pipe\n"); return(0);}
	    cnt+=pipe_n;
	    small[1]='\0';
	    if(small[0]=='@') {
            output[strlen(output)]='\0';
            return strlen(output);
	    }
	    strcat(output, small);
	}
	return 0;
}