#include <openssl/ssl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/timeb.h>

#include "openssl/bn.h"
#include <gmp.h>
#include "openssl/rsa.h"
#include "openssl/dsa.h"
#include "openssl/dh.h"

#include "parameters.h"


#define START 0
#define STOP 1

typedef struct gRSA {
    mpz_t n;              // public modulus
    mpz_t e;              // public exponent
    mpz_t d;              // private exponent
    mpz_t p;              // secret prime factor
    mpz_t q;              // secret prime factor
    mpz_t dmp1;           // d mod (p-1)
    mpz_t dmq1;           // d mod (q-1)
    mpz_t iqmp;           // q^-1 mod p
} gRSA;

void BN2mpz(mpz_t s, const BIGNUM *bn);
void load_pub_key(RSA *params, gRSA *rsa);
void load_keys(RSA *params, gRSA *rsa);
void FDH(BIGNUM *fdh, BIGNUM *tohash);
int sign_crt_gmp(mpz_t *sig, mpz_t *msg, gRSA *rsa);
void rand_num(BIGNUM *x, BIGNUM *n);
double F_Time(int s);
double M_Time(int s, struct timeb *tstart, struct timeb *tend);
double F_Time_Tot(int s);
void hash_mpz_to_str(mpz_t *input,char *);
void FDH_str_to_mpz(const unsigned char *input, int len, mpz_t *output);
int read_from_socket(int socket, int len, char *output);
int write_to_socket(int socket, int len, char *input);
int write_to_pipe(int pipe, char *input, int len);
int read_from_pipe(int pipe, char *output, int max_len);
